
/* lcau25.c				*/
/* (2,5) Linear Cellular Automaton	*/

/* Reference:				*/
/*					*/
/*    Kenneth E. Perry			*/
/*    Abstract Mathematical Art		*/
/*    BYTE				*/
/*    December, 1986			*/
/*    pages 181-192			*/

/*    Copyright (C) 1987		*/
/*    Copyright (C) 1988		*/
/*    Copyright (C) 1989		*/
/*    Harold V. McIntosh		*/
/*    Gerardo Cisneros S.		*/

/* G. Cisneros, 4.3.87						*/
/* 10 April 1987 - modified for (4,2) [HVM]			*/
/* 26 April 1987 - Multiple menus [HVM]				*/
/* 28 April 1987 - back modified to (4,1) [HVM]			*/
/* 28 April 1987 - version for XVI Feria de Puebla [HVM]	*/
/* 14 May 1987 - modified for (3,1) and general rule [HVM]	*/
/* 19 May 1987 - modified for (2,2) [HVM]			*/
/* 20 May 1987 - modified for (2,3) [HVM]			*/
/* 21 May 1987 - Wolfram rule number [HVM]			*/
/* 8 June 1987 - general rule for (4,1) [HVM]		 	*/
/* 12 June 1987 - cartesian product of (2,1) rules [HVM]	*/
/* 12 June 1987 - (2,1) rule with memory  [HVM]			*/
/* 14 June 1987 - individual cycles of evolution [HVM]		*/
/* 17 June 1987 - p-adic representation in plane [HVM]		*/
/* 22 June 1987 - 2 speed gliders via 16 transitions [HVM]	*/
/* 26 June 1987 - push, pop the rule [HVM]			*/
/* 26 June 1987 - conserve position of rule cursor [HVM]	*/
/* 27 June 1987 - incremental rule construction [HVM]		*/
/* 29 June 1987 - conserve position of cell pointer [HVM]	*/
/* 30 June 1987 - mark & unmark transitions, xchg x&X [HVM]	*/
/* 25 July 1987 - display and edit de Bruijn diagrams [HVM]	*/
/* 27 July 1987 - graph of transition probabilities [HVM]	*/
/* 4 September 1987 - PROB21.C for option 't' [HVM]		*/
/* 21 October 1987 - program disks disappeared			*/
/* 20 December 1987 - program reconstructed from listings	*/
/* 20 February 1988 - RIJN21.C for option 'd' [HVM]		*/
/* 11 March 1989 - extended to fourth neighbors [HVM]		*/
/* 11 March 1989 - S option for semitotalistic rules [HVM]	*/
/* 11 March 1989 - x option in srule and trule [HVM]		*/
/* 16 March 1989 - *+- option to change screen color [HVM]	*/
/* 24 September 1989 - edline allows rule design [HVM]		*/

/* # include <bdos.h> */
# include "lcau25.h"
# include "rec.h"
# include "rectbl.h"
# include "keybrd.h"

# define LROW		11	/* first row of line panel */

char xrule[NX][KK][TS-1]= {

	"00010110001", "01101001110",
	"11010011010", "01011111000",
	"10110011011", "10010110110",
	"00110110101", "10011001001"

	};

char   trule[TS];
char   srule[KK][TS-1];
char   ascrule[KK][KK][KK][KK][KK][KK][KK][KK][KK][KK][KK];
char   auxrule[KK][KK][KK][KK][KK][KK][KK][KK][KK][KK][KK];
char   binrule[KK][KK][KK][KK][KK][KK][KK][KK][KK][KK][KK];
char   arr1[AL], arr2[AL];
int    ru0, ru1, ru2, ru3, ru4, ru5, ru6, ru7, ru8, ru9, rua;	/* marker in edrule        */
int    ru0, ru1, ru2, ru3, ru4, ru5, ru6, ru7, ru8;	/* marker in edrule        */
int    co, dn;						/* connectivity, dimension */
int    pr;						/* probability, in mils    */
int    ix0, iy0;					/* origin for pen moves    */
int    dij, di, dj;					/* demo pointer            */
int    rij, rdi, rdj;					/* rec demo pointers 	   */
int    li, lj;						/* cursor for edline       */
int    pr, pl;						/* rule, line prob, mils   */
int    rp, rq;						/* source code in cstr	   */
int    bi;						/* bernstein index	   */
int    ct0, ct1, ct2, ct3;				/* general counters	   */
int    sw0, sw1, sw2, sw3;				/* general switches	   */
int    ex, wy;						/* hiplot coordinates	   */
double nu0, nu1, nu2, nu3;				/* general numeric values  */
double mf;						/* mouse increment	   */
double sm[TS];						/* selfconsistent nbrhd    */
double sn[TS];						/* selfconsistent nbrhd    */
double bx[BD];						/* bernstein factor        */
double bp[BD];
double pv[10];						/* polynomial vector 	   */
double fpa, fpb, fpc;					/* fixed points		   */

Inst   rprg[384];			/* rec program array */
char   cstr[CLEN];			/* console rec program */
char   cprg[CLEN];			/* edit stored program */

extern char recrule[NY][48];

long int r_intpar;

main() {
int  more='r', sc=0;
int  i, j, i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, ia;
char c;

for (i=0; i<TS; i++) trule[i]='0';
for (i=0; i<KK; i++) for (j=0; j<TS-1; j++) srule[i][j]='0';

ru0=0; ru1=0; ru2=0; ru3=0; ru4=0, ru5=0; ru6=0; ru7=0; ru8=0; ru9=0; rua=0;
li=SL/2; lj=0;
ct0=ct1=ct2=ct3=0;					/* counters	       */
sw0=sw1=sw2=sw3=0;					/* switches	       */
co=1;							/* connectivity      */
dn=1;							/* dimension         */
rij=2; rdi=0; rdj=1;
rp=rq=0; cstr[rp]=0;
pr=500;	pl=500;						/* probability, mils */
mf=0.05;
for (i=0; i<BD; i++) bx[i]=1.0;				/* bernstein factor */
bi=0;							/* bernstein index     */

    videopalette(WHCYMAG);				/* white/cyan/magenta */

    tuto();
    while (!kbdst()) rand();				/* wait for keypress */
    kbdin();						/* ignore it */
    videomode(T80X25);
    videoscroll(0,0,CR,71,0,3);				/* menu on blue background */
    videoscroll(CL+8,0,24,71,0,3);
    i=(rand()%NX);
    di=i%DROWS;
    dj=(i/DROWS==1?DCOLS:1);
    xrtosr(i);
    sttoasc();
    auxblnk();						/* clear auxlin */
    ranbchay(arr1,AL,pl);				/* random initial array */

    while (more!='n') {					/* execute multiple runs */
    rmenu();
    lmenu();
    while (0<1) {					/* set up one run */
    c=kbdin();
    if (c=='g') break;					/* go draw graph */
    if (c=='q') more='n';				/* quit for good */
    if (more=='n') break;
    switch (c) {


        case Kf1:					/* F1 = fetch stored rule */
	    dij=dmenu(Kf1);
	    xrtosr(lim(1,dij+1,NX)-1);
	    sttoasc();
	    rmenu();
	    lmenu();
	    break;
        case Kf2:					/* F2 = rec program */
	    rij=lim(1,recdem(Kf2),NY)-1;
	    for (i=0; i<CLEN; i++) cstr[i]=recrule[rij][i];
	    rmenu();
	    lmenu();
	    break;
        case Kf3:					/* f3 = edit rec program */
	    edrec(cstr); 
	    rmenu();
	    lmenu();
	    break;
        case Kf4:					/* f4 = execute rec program */
	    if (rec_c(cstr,rprg,PLEN,dtbl)==1) rec_x(rprg);
	    break;
        case Kf5:					/* f5 = execute demo program */
	    if (rec_c(&recrule[rdi],rprg,PLEN,dtbl)==1) rec_x(rprg);
	    break;

	case '*': sc=2; break;				/* white on red */
	case '+': sc=1; break;				/* white screen */
	case '-': sc=0; break;				/* blue screen */
	case '@':					/* numbered tot rule */
	    nutoto(numin(0));
	    totoasc();
	    rmenu();
	    videocursor(0,CM,0);
	    break;
	case '$':					/* dozen totalistics */
	    j=numin(0);
	    for (i=0; i<12; i++) {
	      nutoto(j+i);
	      totoasc();
	      ranbchay(arr1,AL,pl);
	      evolve(0);
	      }
            videomode(T80X25);
	    rmenu();
	    lmenu();
	    break;
        case 'o': co=numin(0); break;			/* normal connectivity */
	case 'S':					/* semitotalistic rule */
	    xblnk();
	    smenu();
	    edsrule();
	    sttoasc();
	    videocursor(0,CM,0);
	    rmenu();
	    break;
	case 'T':					/* totalistic rule */
	    xblnk();
	    tmenu();
	    edtrule();
	    totoasc();
	    videocursor(0,CM,0);
	    rmenu();
	    break;
	case 't':					/* probability calculation */
	    edtri();
	    rmenu();
	    lmenu();
	    break;
        case 'r':					/* edit rule */	
	    xblnk();
	    edrule();
	    videocursor(0,CM,0);
	    rmenu();
	    break;
        case 'l':					/* edit cell string */
	    xblnk();
	    edline(8,40);
	    videocursor(0,CM,0);
	    lmenu();
	    break;
        case '#':					/* read stored rule */
	    xrtosr(lim(1,dmenu(),NX)-1);
	    sttoasc();
	    rmenu();
	    lmenu();
	    break;
	case 'D':					/* run through samples */
	    for (i=0; i<NX; i++) {
	      xmenu(i);
	      xrtosr(i);
	      sttoasc();
	      ranbchay(arr1,AL,pl);
	      evolve(0);
	      }
	    videomode(T80X25);
	    rmenu();
	    break;
	case 'd':					/* de Bruijn diagram */
	    edijn();
	    rmenu();
	    lmenu();
	    break;
        case 'u':					/* sparse init arry */
	    xblnk();
	    for (i=0; i<AL; i++) arr1[i]=0;
	    arr1[AL/4]=1;
            arr1[AL/2]=1;
            arr1[(3*AL)/4]=1;
            arr1[(3*AL)/4+2]=1;
	    lmenu();
            break;
	case 'w':					/* Wolfram rule # */
	    i=numin(0);
	    wmenu(i);
	    for (i0=0; i0<KK; i0++)
	    for (i1=0; i1<KK; i1++)
	    for (i2=0; i2<KK; i2++)
	    for (i3=0; i3<KK; i3++)
	    for (i4=0; i4<KK; i4++)
	    for (i5=0; i5<KK; i5++)
	    for (i6=0; i6<KK; i6++)
	    for (i7=0; i7<KK; i7++)
	    for (i8=0; i8<KK; i8++)
	    for (i9=0; i9<KK; i9++)
	    for (ia=0; ia<KK; ia++) {
	      ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia]='0'+i%KK;
	      i/=KK;
	    }
	    rmenu();
	    break;
	case 'x':					/* random rule */
	    xblnk();
	    ranpchay(ascrule,auxrule,DS,pr);
	    rmenu();
	    break;
	case 'X':					/* random rule */
	    xblnk();
	    ranchay(ascrule,DS,pr);
	    rmenu();
	    break;
	case 'y':					/* random line */
	    xblnk();
	    ranbchay(arr1,AL,pl);
            lmenu();
	    break;
	case 'Y':					/* symmetrize rule */
	    for (i0=0; i0<KK; i0++)
	    for (i1=0; i1<KK; i1++)
	    for (i2=0; i2<KK; i2++)
	    for (i3=0; i3<KK; i3++)
	    for (i4=0; i4<KK; i4++)
	    for (i5=0; i5<KK; i5++)
	    for (i6=0; i6<KK; i6++)
	    for (i7=0; i7<KK; i7++)
	    for (i8=0; i8<KK; i8++)
	    for (i9=0; i9<KK; i9++)
	    for (ia=0; ia<KK; ia++)
	      ascrule[ia][i9][i8][i7][i6][i5][i4][i3][i2][i1][i0]=
		ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia];
	    break;
	case '.':
	    asctobin();
	    onegen(AL);
	    lmenu();
	    break;
	case '=':
	    for (i=1; i<8;  i++) {
	    for (j=0; j<40; j++) arr1[40*i+j]=arr1[j];}
	    lmenu();
	    break;
	case '~':
	    for (i=1; i<16;  i++) {
	    for (j=0; j<20; j++) arr1[20*i+j]=arr1[j];}
	    lmenu();
	    break;
        default: break;
        }
    }
    if (more=='n') break;
    do {
    evolve(sc);
    videocursor(0,0,0);
    scrstr("?");
    videocursor(0,0,34);
    scrstr("y/n/cr");
    more=kbdin();
    } while (more=='\015');
    videomode(T80X25);					/* reset the screen */
    if (more=='n') break;
    }
  videomode(T80X25);}    

/* edit the rule */
edrule() {
char c;
int  i, i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, ia;

i=6; i0=0; i1=0, i2=0; i3=0; i4=0; i5=0; i6=0; i7=0; i8=0; i9=0; ia=0;
    while (0<1) {
        videocursor(0,CR,i);
        c=kbdin();
        if (c == '\015') break;				/* carriage return exits */
        switch (c) {
        case '0':  case '1':				/* state */
	    ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia] = c;
    	    ia++;
	    if (ia>KK-1) {ia=0; ia++;}
	    if (i9>KK-1) {i9=0; i9++;}
	    if (i8>KK-1) {i8=0; i8++;}
	    if (i7>KK-1) {i7=0; i7++;}
	    if (i6>KK-1) {i6=0; i6++;}
	    if (i5>KK-1) {i5=0; i5++;}
	    if (i4>KK-1) {i4=0; i3++;}
	    if (i3>KK-1) {i3=0; i2++;}
	    if (i2>KK-1) {i2=0; i1++;}
	    if (i1>KK-1) {i1=0; i0++;}
	    if (i0>KK-1) {i2=0; }
            videocattr(0,c,3,1);
            if (i<6+DS) i++;
            break;
        case ' ': case '\315':				/* space = advance */
    	    ia++;
	    if (ia>KK-1) {ia=0; ia++;}
	    if (i9>KK-1) {i9=0; i9++;}
	    if (i8>KK-1) {i8=0; i8++;}
	    if (i7>KK-1) {i7=0; i7++;}
	    if (i6>KK-1) {i6=0; i6++;}
	    if (i5>KK-1) {i5=0; i5++;}
	    if (i4>KK-1) {i4=0; i3++;}
	    if (i3>KK-1) {i3=0; i2++;}
	    if (i2>KK-1) {i2=0; i1++;}
	    if (i1>KK-1) {i1=0; i0++;}
	    if (i0>KK-1) {i2=0; }
            if (i<6+DS) i++;
            break;
        case '\010': case '\313':			/* backspace */
	    if (ia!=0) ia--; else {ia=KK-1;
	    if (i9!=0) i9--; else {i9=KK-1;
	    if (i8!=0) i8--; else {i8=KK-1;
	    if (i7!=0) i7--; else {i7=KK-1;
	    if (i6!=0) i6--; else {i6=KK-1;
	    if (i5!=0) i5--; else {i5=KK-1;
	    if (i4!=0) i4--; else {i4=KK-1;
	    if (i3!=0) i3--; else {i3=KK-1;
	    if (i2!=0) i2--; else {i2=KK-1;
	    if (i1!=0) i1--; else {i1=KK-1;
	    if (i0!=0) i0--; else {i0=KK-1;
	    }}}}}}}}}}}
            if (i>6) i--;
            break;
    case '\013': case '\310':
	auxrule[ru0][ru1][ru2][ru3][ru4][ru5][ru6][ru7][ru8][ru9][rua]='+'; break; /* up arrow */
    case '\012': case '\320':
	auxrule[ru0][ru1][ru2][ru3][ru4][ru5][ru6][ru7][ru8][ru9][rua]=' '; break; /* down arrow */
	default: break;
        }
    }
}

/* edit semitotalistic rule */
edsrule() {char c; int  i, j, k;
for (j=0; j<KK; j++) {
  i=0;
  while (i<TS-1) {
    c=srule[j][i];
    videocursor(0,CR-2+j,57+i);
    videocattr(0,c,3,1);
    c=kbdin();
    if (c=='\015') break;
    switch (c) {
      case '0': case '1':			/* state */
        srule[j][i]=c;
        videocattr(0,c,4,1);
        i++;
        break;
      case ' ': case '\315':			/* space = advance */
        i++;
        break;
      case '\010': case '\313':			/* backspace */
        if (i!=0) i--;
	break;
      case '\013': case '\310': if (j==0) j=KK; j--; break; /* up arrow */
      case '\012': case '\320': j++; if (j==KK) j=0; break; /* down arrow */
      case 'x':					/* random selection */
	for (i=0; i<TS-1; i++) {
	  if (i%8==0) k=rand();
	  srule[j][i]='0'+k%KK; k/=KK;}
	i=0;
	sscrrul(j);
	break;
      default: break;
      } /* end switch */
  }     /* end while  */
}       /* end for    */
}

/* edit totalistic rule */
edtrule() {char c; int  i, k;
  i=0;
  while (i<TS) {
    c=trule[i];
    videocursor(0,3,56+i);
    videocattr(0,c,3,1);
    c=kbdin();
    if (c=='\015') break;
    switch (c) {
      case '0': case '1':				/* state */
        trule[i]=c;
        videocattr(0,c,4,1);
        i++;
        break;
      case ' ': case '\315':			/* space = advance */
        i++;
        break;
      case '\010': case '\313':			/* backspace */
        if (i!=0) i--;
	break;
      case 'x':					/* random selection */
	for (i=0; i<TS; i++) {
	  if (i%8==0) k=rand();
	  trule[i]='0'+k%KK; k/=KK;}
	i=0;
	tscrrul();
	break;
      default: break;
      }
  }
}

/* edit the line of cells */
edline(nr,nc) int nr, nc; {
char c;
int  i, j, k, l, ll, m;
int  aa, bb, cc, dd, ee, ff, gg, hh, ii;
int  rr, ss, tt, uu, vv, ww, xx, yy, zz;

videocursor(0,20,0);
scrstr("insert states with 0, 1; move cursor");
videocursor(0,21,0);
scrstr("with keyboard arrows, space or backspace.");
videocursor(0,22,0);
scrstr("carriage return exits");
while (0<1) {
  m=0;					/* 0/1 = refresh/keep line */
  ll=lj;
  videocursor(0,lj+LROW,li);
  c=kbdin();
  if (c == '\015') {videoscroll(20,0,24,70,0,3); break;}
  switch (c) {
  case '0': case '1':					/* enter state */
    arr1[li+nc*lj]=c-'0';
    li++; if (li==nc) {li=0; lj++;} if (lj==nr) lj=0;
    break;
  case '\012': case '\320':				/* downwards */
    lj++; if (lj==nr) lj=0;
    break;
  case '\013': case '\310':				/* upwards */
    if (lj>0) lj--; else lj=nr-1;
    break;
  case '\014': case '\315': case ' ':			/* forward */
    li++; if (li==nc) {li=0; lj++;} if (lj==nr) lj=0;
    break;
  case '\010': case '\313':				/* backward */
    if (li>0) li--;
      else {li=nc-1; if (lj>0) lj--; else lj=nr-1;}
    break;
  case '<':						/* left margin */
    li=0;
    break;
  case '>':						/* right margin */
    li=nc-1;
    break;
  case 'z':						/* clear single row */
    for (k=0; k<nc; k++) arr1[k+nc*lj]=0;
    lj++; if (lj==nr) lj=0;
    break;
  case 'Z':						/* clear full array */
    for (k=0; k<AL; k++) arr1[k]=0;
    lmenu();
    break;
  case '^':						/* copy line above */
    for (k=0; k<nc; k++)
      arr1[nc*lj+k]=arr1[nc*(lj-1)+k];
    lj++; if (lj==nr) lj=0;
    break;
  case 'x':						/* unmark transitions */
    auxblnk();
    m=1;
    break;				
  case 'q':						/* clean up the area */
    ulmenu(4);
    ulmenu(3);
    m=1;
    break;
  default: break;
      } /* end switch */
  if (m==0) for (k=0; k<nc; k++) {
    videocursor(0,ll+LROW,k);
    l=arr1[k+nc*ll]+'0';
    videoputc(l=='0'?'.':l,1);
    }
  }     /* end while */
}       /* end edline */

/* display a screen of evolution */
evolve(sc) int sc; {
int i, j;
  videomode(COLGRAF);					/* erase the screen */
  videocursor(0,0,0);					/* top text line */
  scrstr(":");
  scrrul(); 
  asctobin();
  for (j=8; j<200; j++) videodot(j,YL-1,2);		/* marginal dots */
  for (j=8; j<200; j++) {				/* evolve for 192 generations */
    switch (sc) {
      case 0: for (i=0; i<YL; i++) videodot(j,i,arr1[i]); break;
      case 1: for (i=0; i<YL; i++) videodot(j,i,arr1[i]*3); break;
      case 2: for (i=0; i<YL; i++) videodot(j,i,arr1[i]+2); break;
      default: break; }
    onegen(AL);
    if (kbdst()) {kbdin(); break;}
    }
}

/* copy ascrule over to binrul */
asctobin() {int i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, ia;
    for (i0=0; i0<KK; i0++)
    for (i1=0; i1<KK; i1++)
    for (i2=0; i2<KK; i2++)
    for (i3=0; i3<KK; i3++)
    for (i4=0; i4<KK; i4++)
    for (i5=0; i5<KK; i5++)
    for (i6=0; i6<KK; i6++)
    for (i7=0; i7<KK; i7++)
    for (i8=0; i8<KK; i8++)
    for (i9=0; i9<KK; i9++)
    for (ia=0; ia<KK; ia++)
      binrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia]=
	ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia]-'0';
}

/* evolution for one generation */

onegen(j) int j; {int i, k, i0, i1, i2, i3, i4; long int ii;
switch (dn) {

case 1:
if (j>10) {
  arr2[0]=binrule[arr1[j-5]][arr1[j-4]][arr1[j-3]][arr1[j-2]][arr1[j-1]]
	[arr1[0]][arr1[1]][arr1[2]][arr1[3]][arr1[4]][arr1[5]];
  arr2[1]=binrule[arr1[j-4]][arr1[j-3]][arr1[j-2]][arr1[j-1]][arr1[0]]
	[arr1[1]][arr1[2]][arr1[3]][arr1[4]][arr1[5]][arr1[6]];
  arr2[2]=binrule[arr1[j-3]][arr1[j-2]][arr1[j-1]][arr1[0]][arr1[1]]
	[arr1[2]][arr1[3]][arr1[4]][arr1[5]][arr1[6]][arr1[7]];
  arr2[3]=binrule[arr1[j-2]][arr1[j-1]][arr1[0]][arr1[1]][arr1[2]]
	[arr1[3]][arr1[4]][arr1[5]][arr1[6]][arr1[7]][arr1[8]];
  arr2[4]=binrule[arr1[j-1]][arr1[0]][arr1[1]][arr1[2]][arr1[3]]
	[arr1[4]][arr1[5]][arr1[6]][arr1[7]][arr1[8]][arr1[9]];
  for (i=5; i<j-5; i++) arr2[i]=binrule[arr1[i-5]][arr1[i-4]][arr1[i-3]][arr1[i-2]]
	[arr1[i-1]][arr1[i]][arr1[i+1]][arr1[i+2]][arr1[i+3]][arr1[i+4]][arr1[i+5]];
  arr2[j-5]=binrule[arr1[j-10]][arr1[j-9]][arr1[j-8]][arr1[j-7]][arr1[j-6]][arr1[j-5]]
	[arr1[j-4]][arr1[j-3]][arr1[j-2]][arr1[j-1]][arr1[0]];
  arr2[j-4]=binrule[arr1[j-9]][arr1[j-8]][arr1[j-7]][arr1[j-6]][arr1[j-5]][arr1[j-4]]
	[arr1[j-3]][arr1[j-2]][arr1[j-1]][arr1[0]][arr1[1]];
  arr2[j-3]=binrule[arr1[j-8]][arr1[j-7]][arr1[j-6]][arr1[j-5]][arr1[j-4]][arr1[j-3]]
	[arr1[j-2]][arr1[j-1]][arr1[0]][arr1[1]][arr1[2]];
  arr2[j-2]=binrule[arr1[j-7]][arr1[j-6]][arr1[j-5]][arr1[j-4]][arr1[j-3]][arr1[j-2]]
	[arr1[j-1]][arr1[0]][arr1[1]][arr1[2]][arr1[3]];
  arr2[j-1]=binrule[arr1[j-6]][arr1[j-5]][arr1[j-4]][arr1[j-3]][arr1[j-2]][arr1[j-1]]
	[arr1[0]][arr1[1]][arr1[2]][arr1[3]][arr1[4]];
} else if (j==8) {
} else if (j==7) {
} else if (j==6) {
} else if (j==5) {
} else if (j==4) {
} else if (j==3) {
} else if (j==2) {
} else if (j==1) {
}
if (co==1) for (ii=0; ii<j; ii++) arr1[(int)ii]=arr2[(int)(ii)];
if (co==0) for (ii=0; ii<j; ii++) arr1[(int)ii]=arr2[(int)((7*ii)%j)];
if (co==2) for (ii=0; ii<j; ii+=2) {arr1[(int)ii]=arr2[(int)((7*ii)%j)];
				 arr1[(int)ii+1]=arr2[(int)((7*ii+1)%j)];}
if (co==3) for (ii=0; ii<j; ii+=3) {arr1[(int)ii]=arr2[(int)((7*ii)%j)];
				 arr1[(int)ii+1]=arr2[(int)((7*ii+1)%j)];
				 arr1[(int)ii+2]=arr2[(int)((7*ii+2)%j)];}
if (co==4) for (ii=0; ii<j; ii+=4) {arr1[(int)ii]=arr2[(int)((7*ii)%j)];
				 arr1[(int)ii+1]=arr2[(int)((7*ii+1)%j)];
				 arr1[(int)ii+2]=arr2[(int)((7*ii+2)%j)];
				 arr1[(int)ii+3]=arr2[(int)((7*ii+3)%j)];}

if (co==5) {for (ii=0; ii<j; ii+=2) {arr1[(int)ii]=arr2[(int)((11*ii)%j)];
				 arr1[(int)ii+1]=arr2[(int)((11*ii+1)%j)];}
	    for (ii=0; ii<j; ii+=8) {
		k=arr1[(int)ii];
		arr1[(int)ii]=arr1[(int)((ii+2)%j)];
		arr1[(int)((ii+2)%j)]=k;
		k=arr1[(int)((ii+1)%j)];
		arr1[(int)((ii+1)%j)]=arr1[(int)((ii+4)%j)];
		arr1[(int)((ii+4)%j)]=k;
		k=arr1[(int)((ii+3)%j)];
		arr1[(int)((ii+3)%j)]=arr1[(int)((ii+5)%j)];
		arr1[(int)((ii+5)%j)]=k;
                }
	    }
if (co==6) {for (ii=0; ii<j; ii+=2) {arr1[(int)ii]=arr2[(int)((11*ii)%j)];
				 arr1[(int)ii+1]=arr2[(int)((11*ii+1)%j)];}
	    for (ii=0; ii<j; ii+=16) {
		k=arr1[(int)ii];
		arr1[(int)ii]=arr1[(int)((ii+2)%j)];
		arr1[(int)((ii+2)%j)]=k;
		k=arr1[(int)((ii+1)%j)];
		arr1[(int)((ii+1)%j)]=arr1[(int)((ii+4)%j)];
		arr1[(int)((ii+4)%j)]=k;
		k=arr1[(int)((ii+3)%j)];
		arr1[(int)((ii+3)%j)]=arr1[(int)((ii+5)%j)];
		arr1[(int)((ii+5)%j)]=k;

		k=arr1[(int)((ii+8)%j)];
		arr1[(int)((ii+8)%j)]=arr1[(int)((ii+10)%j)];
		arr1[(int)((ii+10)%j)]=k;
		k=arr1[(int)((ii+9)%j)];
		arr1[(int)((ii+9)%j)]=arr1[(int)((ii+15)%j)];
		arr1[(int)((ii+15)%j)]=k;
		k=arr1[(int)((ii+11)%j)];
		arr1[(int)((ii+11)%j)]=arr1[(int)((ii+13)%j)];
		arr1[(int)((ii+13)%j)]=k;
		k=arr1[(int)((ii+12)%j)];
		arr1[(int)((ii+12)%j)]=arr1[(int)((ii+14)%j)];
		arr1[(int)((ii+14)%j)]=k;
		}
	    } break;

case 5: if (j<16384) break;
	  for (i0=0; i0<8; i0++) for (i1=0; i1<8; i1++) 
	  for (i2=0; i2<8; i2++) for (i3=0; i3<8; i3++) 
	  for (i4=0; i4<4; i4++)
	    arr2[ix5(i0,i1,i2,i3,i4)]=binrule
	      [arr1[ix5(i0-1,i1,i2,i3,i4)]][arr1[ix5(i0+1,i1,i2,i3,i4)]]
	      [arr1[ix5(i0,i1-1,i2,i3,i4)]][arr1[ix5(i0,i1+1,i2,i3,i4)]]
	      [arr1[ix5(i0,i1,i2-1,i3,i4)]][arr1[ix5(i0,i1,i2+1,i3,i4)]]
	      [arr1[ix5(i0,i1,i2,i3-1,i4)]][arr1[ix5(i0,i1,i2,i3+1,i4)]]
	      [arr1[ix5(i0,i1,i2,i3,i4-1)]][arr1[ix5(i0,i1,i2,i3,i4+1)]]
	      [arr1[ix5(i0,i1,i2,i3,i4)]];
	  for (i=0; i<16384; i++) arr1[i]=arr2[i];
	    break;

default: break;}
}

/* index two dimensions */
int ix2(i,j) int i, j; {return (128+i)%128+128*((128+j)%128);}

/* index three dimensions */
int ix3(i,j,k) int i, j, k; {return (32+i)%32+ 32*((32+j)%32+ 32*((16+k)%16 ));}

/* index four dimensions */
int ix4(i,j,k,l) int i, j, k, l; {
return (16+i)%16+ 16*( (8+j)%8+ 8*( (16+k)%16+ 16*( (8+l)%8 )));}

/* index five dimensions */
int ix5(i,j,k,l,m) int i, j, k, l, m; {
return (8+i)%8+ 8*( (8+j)%8+ 8*( (8+k)%8+ 8*( (8+l)%8+ 8*( (4+m)%4 ))));}

/* generate line of cells in arr1 with n mils of 1's */
rprob(n) int n; {int i; for (i=0; i<AL; i++) arr1[i]=(rand()%1000<n)?1:0;}

/* generate a char array of length n with m mils of ascii 1's */
ranchay(a,n,m) char *a; int n, m; {int i, mm;
mm=20*m; for (i=0; i<n; i++) a[i]=(rand()%20000<mm)?'1':'0';}

/* generate a char array of length n with m mils of binary 1's */
ranbchay(a,n,m) char *a; int n, m; {int i, mm;
mm=20*m; for (i=0; i<n; i++) a[i]=(rand()%20000<mm)?1:0;}

/* generate a protected char array of length n with m mils of 1's */
ranpchay(a,b,n,m) char *a, *b; int n, m; {int i, mm;
mm=20*m; for (i=0; i<n; i++) if (b[i]==' ') a[i]=(rand()%20000<mm)?'1':'0';}

/* clear auxrule to blanks */
auxblnk() {int i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, ia;
for (i0=0; i0<KK; i0++)
for (i1=0; i1<KK; i1++)
for (i2=0; i2<KK; i2++)
for (i3=0; i3<KK; i3++)
for (i4=0; i4<KK; i4++)
for (i5=0; i5<KK; i5++)
for (i6=0; i6<KK; i6++)
for (i7=0; i7<KK; i7++)
for (i8=0; i8<KK; i8++)
for (i9=0; i9<KK; i9++)
for (ia=0; ia<KK; ia++)
auxrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia]=' ';
}

/* tutorial and Help screen */
tuto() {
    videomode(T80X25);
    videocursor(0,2,0);
    scrstr("<Copyright (C) 1987, 1988, 1989 - H.V. McIntosh, G. Cisneros S.>");
    videocursor(0,4,0);
    scrstr("         *********** LIFE in One Dimension ***********");
    videocursor(0,6,0);
    scrstr("Two States - Black(0), Cyan(1).");
    videocursor(0,8,0);
    scrstr("Fifth neighbors - five on each side, eleven altogether.");
    videocursor(0,10,0);
    scrstr("Complete transition rule - random, edited, or stored.");
    videocursor(0,12,0);
    scrstr("Initial Cellular Array - random, edited, or patterned.");
    videocursor(0,14,0);
    scrstr("Submenus in options t(probabilities) and d(de Bruijn)");
    videocursor(0,15,0);
    scrstr("will be displayed in response to typing ?");
    videocursor(0,17,0);
    scrstr("Use any key to terminate a display in progress.");
    videocursor(0,21,0);
    scrstr("Now, ... press any key to continue.");
}

/* rule menu */
rmenu() {
    videocursor(0,CR-5,0);
    scrstr("      00000000000000001111111111111111");
    videocursor(0,CR-4,0);
    scrstr("      00000000111111110000000011111111");
    videocursor(0,CR-3,0);
    scrstr("      00001111000011110000111100001111");
    videocursor(0,CR-2,0);
    scrstr("      00110011001100110011001100110011");
    videocursor(0,CR-1,0);
    scrstr("      01010101010101010101010101010101");
    videocursor(0,CR,0);
    scrstr("Rule: ");
    scrrul();
    videocursor(0,CR,42);
    printf("      ");
    videocursor(0,CR,42);
    if (istot()==1) printf("(T-%2d)",totonu(0));
    videocursor(0,CR+1,42);
    printf("           ");
    videocursor(0,CR+1,42);
    if (isstot()==1) printf("(S-%2d,%2d)",stotonu(0,0),stotonu(0,1));
    videocursor(0,CM,0);
    scrstr("    r(rule), l(line), #nn(stored), g(graph), q(quit), t(prob),");
    videocursor(0,CM+1,0);
    scrstr("         x(rand rule), y(rand line), u(unit line), d(deBruijn)");
    videocursor(0,CM+2,0);
    scrstr("         @nn(tot/rule), $nn(12 tot/rules), T(ed tot/rule).");
    videocursor(0,CM,0);
    }

/* semitotalistic rule menu*/
smenu() {
    videocursor(0,CR-3,50); scrstr("       0123456789A");
    videocursor(0,CR-2,50); scrstr("born:  "); sscrrul(0);
    videocursor(0,CR-1,50); scrstr("live:  "); sscrrul(1);
    videocursor(0,CR-2,57);
    }

/* totalistic rule menu*/
tmenu() {
    videocursor(0,CR-3,50); scrstr("      0123456789AB");
    videocursor(0,CR-2,50); scrstr("rule: "); tscrrul();
    videocursor(0,CR-2,56);
    }

/* line menu */
lmenu() {int i, j; char c;
  for (j=0; 40*j<YL; j++) {
    videocursor(0,CL+j,0);
    for (i=0; i<40; i++) {
      c=arr1[40*j+i]+'0';
      videoputc(c=='0'?'.':c,1);
      }
    }
  videocursor(0,CM,0);
  }

/* display Wolfram rule number */
wmenu(n) int n;
{char xx[4]; int i, nn;  
    nn=sprintf(xx,"%3d",n);
    videocursor(0,2,40);
    for (i=0; i<nn; i++) videoputc(xx[i],1);
    videocursor(0,2,40); }

/* display rule number */
xmenu(n) int n;
{char xx[4]; int i, nn;  
    nn=sprintf(xx,"%3d",n);
    videocursor(0,1,40);
    for (i=0; i<nn; i++) videoputc(xx[i],1);
    videocursor(0,1,40); }

/* clear screen space for rule numbers */
xblnk() {
    videocursor(0,CR-4,40);
    scrstr("      ");
    videocursor(0,CR-3,40);
    scrstr("      ");
    videocursor(0,CR,0); }

/* uniform line menu */
ulmenu(l) int l; {int i, j, k;
  for (j=0; j*SL<YL; j++) {
  for (i=0; i<SL;    i++) {
    videocursor(0,LROW+j,i);
    k=arr1[j*SL+i];
    videocattr(0,k==0?'.':k+'0',l,1);
    }}
  }

/* copy saved rule #n into semitotalistic rule */
xrtosr(n) int n; { int i0, i1;
  xmenu(n+1);
  for (i0=0; i0<KK; i0++)
  for (i1=0; i1<TS-1; i1++)
  srule[i0][i1]=xrule[n][i0][i1];		/* random sample rule */
}

/* change semitotalistic rule to general rule */
sttoasc() {int i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, ia;
for (i0=0; i0<KK; i0++)
for (i1=0; i1<KK; i1++)
for (i2=0; i2<KK; i2++)
for (i3=0; i3<KK; i3++)
for (i4=0; i4<KK; i4++)
for (i5=0; i5<KK; i5++)
for (i6=0; i6<KK; i6++)
for (i7=0; i7<KK; i7++)
for (i8=0; i8<KK; i8++)
for (i9=0; i9<KK; i9++)
for (ia=0; ia<KK; ia++)
ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia]
	=srule[i5][i0+i1+i2+i3+i4+i6+i7+i8+i9+ia];
}

/* change totalistic rule to general rule */
totoasc() {int i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, ia;
for (i0=0; i0<KK; i0++)
for (i1=0; i1<KK; i1++)
for (i2=0; i2<KK; i2++)
for (i3=0; i3<KK; i3++)
for (i4=0; i4<KK; i4++)
for (i5=0; i5<KK; i5++)
for (i6=0; i6<KK; i6++)
for (i7=0; i7<KK; i7++)
for (i8=0; i8<KK; i8++)
for (i9=0; i9<KK; i9++)
for (ia=0; ia<KK; ia++)
ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia]
	=trule[i0+i1+i2+i3+i4+i5+i6+i7+i8+i9+ia];
}

/* change decimal totalistic rule to sum values */
nutoto(x) int x; {int i; for (i=0; i<TS; i++) {trule[i]=x%KK+'0'; x/=KK;}}

/* change sum values to decimal semitotalistic rule */
int stotonu(i,j) int i, j; {int r;
 r=0; if (i<TS-1) r=(srule[j][i]-'0')+KK*stotonu(i+1,j);
  return r; }

/* change sum values to decimal totalistic rule */
int totonu(i) int i; {int r;
 r=0; if (i<TS) r=(trule[i]-'0')+KK*totonu(i+1);
  return r; }

/* test whether a rule is semi totalistic */
int isstot() {
char st[KK][TS-1];
int  i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, ia, l;
    l=1;
    st[0][0]=ascrule[0][0][0][0][0][0][0][0][0][0][0];
    st[0][1]=ascrule[1][0][0][0][0][0][0][0][0][0][0];
    st[0][2]=ascrule[1][1][0][0][0][0][0][0][0][0][0];
    st[0][3]=ascrule[1][1][1][0][0][0][0][0][0][0][0];
    st[0][4]=ascrule[1][1][1][1][0][0][0][0][0][0][0];
    st[0][5]=ascrule[1][1][1][1][1][0][0][0][0][0][0];
    st[0][6]=ascrule[1][1][1][1][1][0][1][0][0][0][0];
    st[0][7]=ascrule[1][1][1][1][1][0][1][1][0][0][0];
    st[0][8]=ascrule[1][1][1][1][1][0][1][1][1][0][0];
    st[0][9]=ascrule[1][1][1][1][1][0][1][1][1][1][0];
    st[0][10]=ascrule[1][1][1][1][1][0][1][1][1][1][1];
    st[1][0]=ascrule[0][0][0][0][0][1][0][0][0][0][0];
    st[1][1]=ascrule[1][0][0][0][0][1][0][0][0][0][0];
    st[1][2]=ascrule[1][1][0][0][0][1][0][0][0][0][0];
    st[1][3]=ascrule[1][1][1][0][0][1][0][0][0][0][0];
    st[1][4]=ascrule[1][1][1][1][0][1][0][0][0][0][0];
    st[1][5]=ascrule[1][1][1][1][1][1][0][0][0][0][0];
    st[1][6]=ascrule[1][1][1][1][1][1][1][0][0][0][0];
    st[1][7]=ascrule[1][1][1][1][1][1][1][1][0][0][0];
    st[1][8]=ascrule[1][1][1][1][1][1][1][1][1][0][0];
    st[1][9]=ascrule[1][1][1][1][1][1][1][1][1][1][0];
    st[1][10]=ascrule[1][1][1][1][1][1][1][1][1][1][1];
    for (i0=0; i0<KK; i0++)
    for (i1=0; i1<KK; i1++)
    for (i2=0; i2<KK; i2++)
    for (i3=0; i3<KK; i3++)
    for (i4=0; i4<KK; i4++)
    for (i5=0; i5<KK; i5++)
    for (i6=0; i6<KK; i6++)
    for (i7=0; i7<KK; i7++)
    for (i8=0; i8<KK; i8++)
    for (i9=0; i9<KK; i9++)
    for (ia=0; ia<KK; ia++)
    if (ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia]
	!=st[i5][i0+i1+i2+i3+i4+i6+i7+i8+i9+ia]) l=0;
    return l; }

/* test whether a rule is totalistic */
int istot() {int tr[TS], i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, ia, l;
    l=1;
    tr[0]=ascrule[0][0][0][0][0][0][0][0][0][0][0];
    tr[1]=ascrule[1][0][0][0][0][0][0][0][0][0][0];
    tr[2]=ascrule[1][1][0][0][0][0][0][0][0][0][0];
    tr[3]=ascrule[1][1][1][0][0][0][0][0][0][0][0];
    tr[4]=ascrule[1][1][1][1][0][0][0][0][0][0][0];
    tr[5]=ascrule[1][1][1][1][1][0][0][0][0][0][0];
    tr[6]=ascrule[1][1][1][1][1][1][0][0][0][0][0];
    tr[7]=ascrule[1][1][1][1][1][1][1][0][0][0][0];
    tr[8]=ascrule[1][1][1][1][1][1][1][1][0][0][0];
    tr[9]=ascrule[1][1][1][1][1][1][1][1][1][0][0];
    tr[10]=ascrule[1][1][1][1][1][1][1][1][1][1][0];
    tr[11]=ascrule[1][1][1][1][1][1][1][1][1][1][1];
    for (i0=0; i0<KK; i0++)
    for (i1=0; i1<KK; i1++)
    for (i2=0; i2<KK; i2++)
    for (i3=0; i3<KK; i3++)
    for (i4=0; i4<KK; i4++)
    for (i5=0; i5<KK; i5++)
    for (i6=0; i6<KK; i6++)
    for (i7=0; i7<KK; i7++)
    for (i8=0; i8<KK; i8++)
    for (i9=0; i9<KK; i9++)
    for (ia=0; ia<KK; ia++)
    if (ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia]
	!=tr[i0+i1+i2+i3+i4+i5+i6+i7+i8+i9+ia]) l=0;      
    return l; }

/* limit j to interval (i,k) */
int lim(i,j,k) int i, j, k;
    {if (i>=j) return i; if (k<=j) return k; return j;}

/* display the rule number on the screen */
scrrul() {int i, i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, ia;
  for (i0=0; i0<KK; i0++)
  for (i1=0; i1<KK; i1++)
  for (i2=0; i2<KK; i2++)
  for (i3=0; i3<KK; i3++)
  for (i4=0; i4<KK; i4++) {
    i=0;
    for (i5=0; i5<KK; i5++)
    for (i6=0; i6<KK; i6++)
    for (i7=0; i7<KK; i7++)
    for (i8=0; i8<KK; i8++)
    for (i9=0; i9<KK; i9++)
    for (ia=0; ia<KK; ia++)
      i=KK*i+ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia]-'0';
    videoputc(i%16+'0',1);      
    }
}

/* turn ascrule into its Wolfram number and place on screen at (i,j) */
woruno(i,j) int i, j; {
char xx[24];
int  k, l, i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, ia;
  k=0;
  for (i0=KK-1; i0>=0; i0--)
  for (i1=KK-1; i1>=0; i1--)
  for (i2=KK-1; i2>=0; i2--)
  for (i3=KK-1; i3>=0; i3--)
  for (i4=KK-1; i4>=0; i4--)
  for (i5=KK-1; i5>=0; i5--)
  for (i6=KK-1; i6>=0; i6--)
  for (i7=KK-1; i7>=0; i7--)
  for (i8=KK-1; i8>=0; i8--)
  for (i9=KK-1; i9>=0; i9--)
  for (ia=KK-1; ia>=0; ia--)
    k=KK*k+(ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia]-'0');
  k=sprintf(xx,"%10d",k);
  videocursor(0,i,j);
  for (l=0; l<k; l++) videoputc(xx[l],1);
}

txtmode() {videomode(T80X25); rmenu(); lmenu();}
grfmode() {videomode(COLGRAF);}

/* display semitotalistic rule number by sum */
sscrrul(j) int j; {int i; for (i=0; i<TS-1; i++) videoputc(srule[j][i],1);}

/* display totalistic rule number by sum */
tscrrul() {int i; for (i=0; i<TS; i++) videoputc(trule[i],1);}

/* write a string in graphics mode */
scrstr(s) char *s; {for (; *s != '\0'; s++) videoputc(*s,1);}

/* place time and date on screen */ 
tidat() {long cl[1]; char *ctime(); time(cl); printf("%26s",ctime(cl));}

/* keyboard status */
int kbdst() {return(bdos(11) & 0xFF);}

/* direct keyboard input, no echo */
kbdin() {int c;
  if ((c = bdos(7) & 0xFF) == '\0') c = (bdos(7) & 0xFF) | 0x80;
  return(c);
}

/* turn ASCII character into hexadecimal value */
int nibbl(c) int c; {
  if (c>='0'&&c<='9') return (c-'0');
  if (c>='a'&&c<='f') return (c-'a'+10);
  if (c>='A'&&c<='F') return (c-'A'+10);
  return 0;
}

/* turn a nibble into an ASCII character */
int nibasc(i) int i; {if (i<10) return '0'+i; else return 'A'+i-10;}

/* read number from keyboard */
int numin(n) int n; {char c;
  c=kbdin();
  if (c>='0'&&c<='9') return(numin(10*n+(c-'0'))); else return(n);
}

/* end lcau25.c */
