
/* SQAR21.C					*/
/* Five-Dimensional Square Lattice Life		*/
/* binary cells, first neighbors		*/
/* ---- von Neumann neighborhoods ----		*/
/*						*/
/* Reference:					*/
/*						*/
/* Harold V. McIntosh, 16 May 1987		*/
/* 7 July 1987 - separate rules for each state	*/
/* 8 July 1988 - binary, 3 neighborhood classes	*/
/* 9 July 1988 - hexagonal lattice 		*/
/* 20 July 1988 - von Neumann neighborhgood	*/

# import  "nxlcau25.h"
# import  <dpsclient/psops.h>
# import  <appkit/graphics.h>

# define QS	    32  /* number of symmetry classes		*/
# define NF	    50	/* lines in live frequency graph	*/
# define NG	     6	/* number of Golay Classes		*/

char    golRule[KK*NG];	/* evolutionary rule       */
char    sqrule[DS];	/* evolutionary rule       */
char    qarule[DS];	/* auxiliary rule flags    */
int     li, lj;		/* cursor in cell array    */
int	lf[KK];		/* number of live cells	   */
int     gc[DS];		/* Golay code key array    */
int     tc[DS];		/* totalistic key array    */
double  bp[RR+1];	/* Bernstein coefficients  */

extern char   ascrule[KK][KK][KK][KK][KK][KK][KK][KK][KK][KK][KK];

int     totb[DS];
int     gotb[DS];

mainsquar() {}	   /* end main */

/* graph statistics of rule usage */
/* beware limited count capacity  */
/* large counts look negative     */
arrstat() {int i, j, m, n; double s, t;
for (i=0, s=0.0; i<QS; i++) s+=(double)rus[i];
if  (s<1.0) {for (i=0; i<QS; i++) rus[i]=0; return;};
for (i=0; i<QS; i++) {
  t=333.3*(((double)rus[i])/s);	/* full scale = 1/10 population */
  if (t>=32000.0) t=50.0;	/* larger looks negative */
  n=(int)t;
  arsdot(i,33,0);
  m=33; if (n>33) {n=33; m=34;};
  if (n>0) for (j=0; j<n; j++) arsdot(i,j,1);
  for (j=n; j<m; j++) arsdot(i,j,2);
  rus[i]=0;
  }
}

/* graph fraction of live cells */
/* y = column for dot */
/* n = value to be graphed */
arlstat(y,n) int y, n; {int i, k;
  k=(int)(100.0*(((double)(n))/((double)(AL*AL))));
  if (k<=0) k=1; if (k>=100) k=99;
  for (i=0; i<k; i++) arldot(i,y,1);
  for (i=k; i<101; i++) arldot(i,y,2);
  for (i=0; i<101; i++) arldot(i,y+1,0);
  for (i=0; i<101; i+=10) arldot(i,y+1,3);
}

/* generate the Golay code table */
argolta() {
int i, j, k;
int i0, i1, i2, i3;
int arbinq();
k=0;
for (i3=0; i3<KK; i3++)
for (i2=0; i2<KK; i2++)
for (i1=0; i1<KK; i1++)
for (i0=0; i0<KK; i0++) {
  i=arbinq(i0,i1,i2,i3);
  j=arbinq(i1,i2,i3,i0);
  if (j<i) i=j;
  j=arbinq(i2,i3,i0,i1);
  if (j<i) i=j;
  j=arbinq(i3,i0,i1,i2);
  if (j<i) i=j;
  j=arbinq(i0,i3,i2,i1);
  if (j<i) i=j;
  j=arbinq(i3,i2,i1,i0);
  if (j<i) i=j;
  j=arbinq(i2,i1,i0,i3);
  if (j<i) i=j;
  j=arbinq(i1,i0,i3,i2);
  if (j<i) i=j;
  j=arbinq(i0,i1,i2,i3);
  if (i==j) {gc[i]=k; k++;} else {gc[j]=gc[i];};
  }
/* printf("no. cl. = %3d",k); */
}

/* calculate binary equivalent */
int arbinq(i0,i1,i2,i3) int i0, i1, i2, i3; {return i0+KK*(i1+KK*(i2+KK*i3));}

/* calculate Golay class of a cell */
/* no check is made that cell isn't at array edge */
int argolcl(i0,i1,arr1) int  i0, i1, arr1[AL][AL]; {int gn;
gn=arr1[i0-1][i1];
gn=KK*gn+arr1[i0][i1-1];
gn=KK*gn+arr1[i0+1][i1];
gn=KK*gn+arr1[i0][i1+1];
if (arr1[i0][i1]==1) gn+=GH;
gn=gc[gn];
return gn;
}

/* plot self-consistent probability graph */
arbpanel() {arburnc(); argfram(100); argdiag(100); bargraf(RR,100);}

/* generate Bernstein coefficients */
arburnc() {int i;
  for (i=0; i<RR+1; i++) bp[i]=0.0;
  if (golRule[0]=='1') bp[0]+=1.0;
  if (golRule[1]=='1') bp[1]+=4.0;
  if (golRule[2]=='1') bp[2]+=4.0;
  if (golRule[3]=='1') bp[2]+=2.0;
  if (golRule[4]=='1') bp[3]+=4.0;
  if (golRule[5]=='1') bp[4]+=1.0;
  if (golRule[6]=='1') bp[1]+=1.0;
  if (golRule[7]=='1') bp[2]+=4.0;
  if (golRule[8]=='1') bp[3]+=4.0;
  if (golRule[9]=='1') bp[3]+=2.0;
  if (golRule[10]=='1') bp[4]+=4.0;
  if (golRule[11]=='1') bp[5]+=1.0;
}

/* arldot(i,j,l) - place dot on live probability  screen at (i,j) in color l */
arldot(i,j,l) int i, j, l; {videodot(90-NF+j,i,l);}

/* arsdot(i,j,l) - place dot on statistics screen at (i,j) in color l */
arsdot(i,j,l) int i, j, l; {videodot(199-10-j,70+5*i,l);}

/* clear the cell array */
zarry(arr1) int  arr1[AL][AL]; {int i0, i1;
for (i0=0; i0<AL; i0++) for (i1=0; i1<AL; i1++) arr1[i0][i1]=0;
}

/* display rule menu */
armenu() {}

// /* calculate wru from an argument rule	*/
// void wrru21(c,ru) char *c, ru[KK][KK][KK]; {int  k, i0, i1, i2;
//   k=0;
//   for (i0=KK-1; i0>=0; i0--)
//   for (i1=KK-1; i1>=0; i1--)
//   for (i2=KK-1; i2>=0; i2--)
//     k=2*k+(ru[i0][i1][i2]-'0');
//   sprintf(c,"%3d",k);
// }

/* display rule number by sum in color */
arscrrul(y,x) int x, y; {int i, n;
n=QS;
videocursor(0,y,x);
  arscrstr("  ");
  for (i=0;   i<n;   i++) arresch(i);
videocursor(0,y,x);
}

/* display rule number as a hexadecimal */
harscrrul() {int i, k;
  videocursor(0,0,0);
  printf("    ");
  for (i=0,  k=0; i<12;  i++) k=KK*k+sqrule[i]-'0';
  printf("%3.3X",k);
  videocursor(0,0,0);
  }

/* display Wolfram rule number of cross section */
warscrrul() {int i;
  i=sqrule[11]-'0';
  i=KK*i+sqrule[10]-'0';
  i=KK*i+sqrule[3]-'0';
  i=KK*i+sqrule[1] -'0';
  i=KK*i+sqrule[10]-'0';
  i=KK*i+sqrule[9]-'0';
  i=KK*i+sqrule[1] -'0';
  i=KK*i+sqrule[0] -'0';
  videocursor(0,0,32);
  printf("     ");
  videocursor(0,0,32);
  printf("[%2d]",i);
  videocursor(0,0,0);
  }

/* edit the rule */
aredrule() {}

/* restore character in edrule */
arresch(i) int i; {videoputc(sqrule[i],(qarule[i]==' ')?1:2);}

/* edit the plane of cells */
aredline(arr1) int arr1[AL][AL]; {}     /* end aredline */

/* generate random cells in arr1 */
ranarr(arr1) int  arr1[AL][AL]; {int i0, i1, i;
for (i=0, i0=0; i0<AL; i0++) for (i1=0; i1<AL; i1++) arr1[i0][i1]=random()%KK;
}

/* generate random cells in the center of arr1 */
rancen(arr1) int  arr1[AL][AL]; {int i0, i1, ia, ib;
  for (i0=0; i0<AL; i0++) 
  for (i1=0; i1<AL; i1++) arr1[i0][i1]=0;
  ia=AL/3; ib=2*(AL/3);
  for (i0=ia; i0<ib; i0++)
  for (i1=ia; i1<ib; i1++) arr1[i0][i1]=random()%KK;
}

/* generate a random rule */
arranrul() {int i; for (i=0; i<QS; i++) sqrule[i]=random()%KK+'0';}

/* write a string in graphics mode */
arscrstr(s) char *s; {for (; *s != '\0'; s++) videoputc(*s,1); }

// # include "nxruin21.c"

/* end nxsqar21.c */
